"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
try:
    from PyQt4.QtCore import pyqtSignal, QObject, pyqtSlot
except ImportError:
    from PyQt5.QtCore import pyqtSignal, QObject, pyqtSlot
from transphire import transphire_utils as tu


class PlotWorker(QObject):
    """
    Plot different information about motion correction and ctf estimation.

    Inherits:
    QObject

    Signals:
    sig_data - Emitted, if data for plotting is found (name|str, data|object, directory|str, settings|object)
    sig_notification - Emitted, if phase plate limit is reached. (text|str)
    """
    sig_data = pyqtSignal(str, object, str, object)
    sig_calculate = pyqtSignal()

    def __init__(self, parent=None):
        """
        Initialise variables.

        Arguments:
        parent - Parent widget (default None)

        Returns:
        None
        """
        super(PlotWorker, self).__init__(parent)
        self.settings = []
        self.sig_calculate.connect(self.calculate_array)

    def reset_list(self):
        self.settings = []

    @pyqtSlot(str, object, object)
    def set_settings(self, name, directory_name, settings):
        """
        Set settings for the calculation of the arrays.

        name - Name of the software that calls the calculation
        directory_name - Name of the directory that contains the log files
        settings - TranSPHIRE settings

        Returns:
        None
        """
        self.settings.append([name, directory_name, settings])
        self.calculate_array_now(
            name=name,
            directory_name=directory_name,
            settings=settings
            )

    @pyqtSlot()
    def calculate_array(self):
        """
        Calculate array.

        Returns:
        None
        """
        for name, directory_name, settings in self.settings:
            self.calculate_array_now(
                name=name,
                directory_name=directory_name,
                settings=settings
                )

    def calculate_array_now(self, name, directory_name, settings):
        if name == 'Later' or name == 'False':
            data = None
        else:
            data, _ = tu.get_function_dict()[name]['plot_data'](
                name=name,
                directory_name=directory_name
                )
        if data is None:
            pass
        elif data.size == 0:
            pass
        else:
            self.sig_data.emit(name, data, directory_name, settings)



"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import re
import collections as co
import numpy as np
from transphire import transphire_utils as tu


def get_motion_default(settings, motion_frames, queue_com, name):
    """
    Set the default values for the motion correction software.

    settings - TranSPHIRE settings.
    motion_frames - Sub frame settings.
    queue_com - Queue for communication.
    name - Name of the process.

    Returns:
    True, if dose weighting will be applied.
    """

    motion_name = settings['Copy']['Motion']
    if motion_name == 'MotionCor2 v1.0.0' or \
            motion_name == 'MotionCor2 v1.0.5' or \
            motion_name == 'MotionCor2 v1.1.0':
        motion_frames['last'] = \
            int(settings['General']['Number of frames']) - \
            int(settings[motion_name]['-Trunc'])
        motion_frames['first'] = \
            int(settings[motion_name]['-Throw']) + 1

        return bool(
                settings[motion_name]['-FmDose'] != '0' and
                settings[motion_name]['-PixSize'] != '0' and
                settings[motion_name]['-kV'] != '0'
                )

    else:
        message = '\n'.join([
            '{0}: Motion version not known.'.format(motion_name),
            'Please contact the TranSPHIRE authors!'
            ])
        queue_com['error'].put(
            message,
            name
            )
        raise IOError(message)


def get_dw_file_name(output_transfer_scratch, file_name, settings, queue_com, name):
    """
    Get the name of the dose weighted file directly after the program finished.

    output_transfer - Name of the folder in the scratch directory.
    file_name - File name of the root_name path.
    settings - TranSPHIRE settings.
    queue_com - Queue for communication.
    name - Name of the process.

    Returns:
    File path of the DW file.
    """
    motion_name = settings['Copy']['Motion']
    if motion_name == 'MotionCor2 v1.0.0' or \
            motion_name == 'MotionCor2 v1.0.5' or \
            motion_name == 'MotionCor2 v1.1.0':
        return os.path.join(
            output_transfer_scratch,
            '{0}_DW.mrc'.format(file_name)
            )

    else:
        message = '\n'.join([
            '{0}: Motion version not known.'.format(motion_name),
            'Please contact the TranSPHIRE authors!'
            ])
        queue_com['error'].put(
            message,
            name
            )
        raise IOError(message)


def get_motion_command(file_input, file_output_scratch, file_log_scratch, settings, queue_com, name, do_subsum):
    """
    Get the command for the selected motion software.

    file_input - Input file for motion correction.
    file_output_scratch - Output file name
    file_log_scratch - Logfile path on the scratch disc
    settings - TranSPHIRE settings.
    queue_com - Queue for communication.
    name - Name of the process.

    Returns:
    Motion command
    """
    motion_name = settings['Copy']['Motion']
    command = None
    block_gpu = None
    gpu_list = None
    if motion_name == 'MotionCor2 v1.0.0' or \
            motion_name == 'MotionCor2 v1.0.5' or \
            motion_name == 'MotionCor2 v1.1.0':
        command, gpu = create_motion_cor_2_v1_0_0_command(
            motion_name=motion_name,
            file_input=file_input,
            file_output=file_output_scratch,
            file_log=file_log_scratch,
            settings=settings,
            queue_com=queue_com,
            name=name,
            do_subsum=do_subsum,
            )
        gpu_list = gpu.split()

        if motion_name == 'MotionCor2 v1.0.0':
            block_gpu = False
        elif motion_name == 'MotionCor2 v1.0.5':
            block_gpu = True
        elif motion_name == 'MotionCor2 v1.1.0':
            if settings[motion_name]['-GpuMemUsage'] == '0':
                block_gpu = False
            else:
                block_gpu = True

    else:
        message = '\n'.join([
            '{0}: Motion version not known.'.format(settings['Copy']['Motion']),
            'Please contact the TranSPHIRE authors!'
            ])
        queue_com['error'].put(
            message,
            name
            )
        raise IOError(message)

    assert command is not None, 'command not specified: {0}'.format(motion_name)
    assert block_gpu is not None, 'block_gpu not specified: {0}'.format(motion_name)
    assert gpu_list is not None, 'gpu_list not specified: {0}'.format(motion_name)

    return command, block_gpu, gpu_list


def create_motion_cor_2_v1_0_0_command(motion_name, file_input, file_output, file_log, settings, queue_com, name, do_subsum):
    """
    Create the MotionCor2 v1.0.0 command

    file_input - Input file for motion correction.
    file_output - Output filename
    file_log - Logfile name
    settings - TranSPHIRE settings.
    queue_com - Queue for communication.
    name - Name of the process.

    Returns:
    Command for MotionCor2 v1.0.0
    """

    ignore_list = []
    ignore_list.append('Split Gpu?')
    ignore_list.append('-Gpu')
    ignore_list.append('dose cutoff')
    command = []
    # Start the program
    command.append('{0}'.format(settings['Path'][motion_name]))
    # Input Micrograph
    _, extension = os.path.splitext(file_input)
    if extension == '.tiff' or \
            extension == '.tif':
        command.append('-InTiff')
        command.append('{0}'.format(file_input))

    elif extension == '.mrc':
        command.append('-InMrc')
        command.append('{0}'.format(file_input))

    else:
        message = '{0}: Not known!'.format(extension)
        queue_com['error'].put(message, name)
        raise IOError(message)

    # Output micrograph
    command.append('-OutMrc')
    command.append('{0}'.format(file_output))
    if do_subsum:
        # Write the output stack
        command.append('-OutStack')
        command.append('1')
    else:
        pass
    # Log file
    command.append('-LogFile')
    command.append('{0}'.format(file_log))

    if settings[motion_name]['Split Gpu?'] == 'True':
        try:
            gpu_id = int(name.split('_')[-1])-1
        except ValueError:
            gpu_id = 0
        try:
            gpu = settings[motion_name]['-Gpu'].split()[gpu_id]
        except IndexError:
            raise UserWarning('There are less gpus provided than threads available! Please restart with the same number of pipeline processors as GPUs provided and restart! Stopping this thread!')
    else:
        gpu = settings[motion_name]['-Gpu']

    command.append('-Gpu')
    command.append('{0}'.format(gpu))

    for key in settings[motion_name]:
        if key in ignore_list:
            continue
        elif settings[motion_name][key]:
            command.append(key)
            command.append(
                '{0}'.format(settings[motion_name][key])
                )
        else:
            continue

    return ' '.join(command), gpu


def create_sum_movie_command(
        motion_frames, file_input, file_output, file_shift, file_frc,
        settings, queue_com, name
        ):
    """
    Create the SumMovie command.

    motion_frames - Sub frames settings dictionary
    file_input - File to sum.
    file_output - Output file name
    file_shift - Output shift file name
    file_frc - Output frc file name
    settings - TranSPHIRE settings
    queue_com - Queue for communication
    name - Name of the process

    Returns:
    Command for SumMovie
    """
    command = create_sum_movie_v1_0_2_command(
        motion_frames=motion_frames,
        file_input=file_input,
        file_output=file_output,
        file_shift=file_shift,
        file_frc=file_frc,
        settings=settings,
        queue_com=queue_com,
        name=name
        )
    block_gpu = False
    gpu_list = []
    return command, block_gpu, gpu_list


def create_sum_movie_v1_0_2_command(
        motion_frames, file_input, file_output, file_shift, file_frc,
        settings, queue_com, name
        ):
    """
    Create the SumMovie v1.0.2 command.

    motion_frames - Sub frames settings dictionary
    file_input - File to sum.
    file_output - Output file name
    file_shift - Output shift file name
    file_frc - Output frc file name
    settings - TranSPHIRE settings
    queue_com - Queue for communication
    name - Name of the process

    Returns:
    Command for Summovie v1.0.2
    """
    sum_movie_command = []
    # Input file
    sum_movie_command.append('{0}'.format(file_input))
    # Number of frames

    motion_name = settings['Copy']['Motion']
    if motion_name == 'MotionCor2 v1.0.0' or \
            motion_name == 'MotionCor2 v1.0.5' or \
            motion_name == 'MotionCor2 v1.1.0':
        sum_movie_command.append('{0}'.format(
            int(settings['General']['Number of frames']) -
            int(settings[motion_name]['-Trunc']) -
            int(settings[motion_name]['-Throw'])
            ))

    else:
        message = '\n'.join([
            '{0}: Motion version not known.'.format(motion_name),
            'Please contact the TranSPHIRE authors!'
            ])
        queue_com['error'].put(
            message,
            name
            )
        raise IOError(message)

    # Output sum file
    sum_movie_command.append('{0}'.format(file_output))
    # Shift file
    sum_movie_command.append('{0}'.format(file_shift))
    # FRC file
    sum_movie_command.append('{0}'.format(file_frc))
    # First frame
    sum_movie_command.append('{0}'.format(motion_frames['first']))
    # Last frame
    sum_movie_command.append('{0}'.format(motion_frames['last']))
    # Pixel size

    if motion_name == 'MotionCor2 v1.0.0' or \
            motion_name == 'MotionCor2 v1.0.5' or \
            motion_name == 'MotionCor2 v1.1.0':
        sum_movie_command.append(
            '{0}'.format(settings[motion_name]['-PixSize'])
            )

    else:
        message = '\n'.join([
            '{0}: Motion version not known.'.format(motion_name),
            'Please contact the TranSPHIRE authors!'
            ])
        queue_com['error'].put(
            message,
            name
            )
        raise IOError(message)

    # Dose correction
    sum_movie_command.append('No')

    command = 'echo "{0}" | {1}'.format(
        '\n'.join(sum_movie_command),
        '{0}'.format(settings['Path']['SumMovie v1.0.2'])
        )

    return command


def combine_motion_outputs(
        data,
        data_original,
        settings,
        queue_com,
        shared_dict,
        name,
        log_file,
        sum_file,
        dw_file,
        ):
    """
    Combine the motion outputs to one micrograph and one relion star file.

    root_path - Root path of the file
    file_name - File name of the ctf file.
    settings - TranSPHIRE settings
    queue_com - Queue for communication
    name - Name of process
    sum_file - Name of the dose uncorrected sum file

    Returns:
    None
    """
    motion_name = settings['Copy']['Motion']
    motion_settings = settings[motion_name]
    log_folder = os.path.dirname(log_file)
    stack_folder = settings['stack_folder']
    project_folder = '{0}/'.format(settings['project_folder'])
    sum_file = sum_file.replace(project_folder, '')
    if dw_file:
        dw_file = dw_file.replace(project_folder, '')
    else:
        pass

    output_name_mic = os.path.join(
        log_folder,
        '{0}_transphire_motion.txt'.format(os.path.basename(sum_file))
        )
    output_name_star = os.path.join(
        log_folder,
        '{0}_transphire_motion.star'.format(os.path.basename(sum_file))
        )
    output_name_star_relion3 = os.path.join(
        log_folder,
        '{0}_transphire_motion_relion3.star'.format(os.path.basename(sum_file))
        )

    output_name_mic_combined = os.path.join(
        project_folder,
        '{0}_transphire_motion.txt'.format(motion_name.replace(' ', '_'))
        )
    output_name_star_combined = os.path.join(
        project_folder,
        '{0}_transphire_motion.star'.format(motion_name.replace(' ', '_'))
        )
    output_name_star_relion3_combined = os.path.join(
        project_folder,
        '{0}_transphire_motion_relion3.star'.format(motion_name.replace(' ', '_'))
        )

    # SPHIRE
    with open(output_name_mic, 'w') as write:
        write.write('{0}\n'.format(os.path.basename(sum_file)))

    # RELION 2
    header_star = co.OrderedDict()
    if dw_file:
        header_star['_rlnMicrographNameNoDW'] = sum_file
        header_star['_rlnMicrographName'] = dw_file
    else:
        header_star['_rlnMicrographName'] = sum_file

    export_lines_star = [get_relion_header(header_star.keys())]
    create_export_data(header_star.values(), export_lines_star)
    with open(output_name_star, 'w') as write:
        write.write('\n'.join(export_lines_star))

    # RELION 3
    with open(log_file, 'r') as read:
        data_read = read.read()

    stack_size = re.search(
        r'^Stack[ ]+size:[ ]+([0-9]+)[ ]+([0-9]+)[ ]+([0-9]+)$',
        data_read,
        re.MULTILINE
        )
    movie_name = re.search( r'^-(?:InTiff|InMrc)[ ]+([^ ]+)$', data_read, re.MULTILINE).group(1)

    if settings['Copy']['Compress data'] == 'True':
        movie_name = movie_name.replace(stack_folder, 'Compress')
        movie_name = movie_name.replace('.mrc', '.tiff')
    else:
        pass

    data_meta = [
        '',
        'data_general',
        '',
        '_rlnImageSizeX {0}'.format(stack_size.group(1)),
        '_rlnImageSizeY {0}'.format(stack_size.group(2)),
        '_rlnImageSizeZ {0}'.format(stack_size.group(3)),
        '_rlnMicrographMovieName {0}'.format(movie_name.replace(project_folder, '')),
        ]
    if motion_settings['-Gain']:
        new_gain = os.path.join(
            project_folder,
            '{0}_gain.mrc'.format(os.path.basename(motion_settings['-Gain']))
            )
        if not os.path.exists(new_gain):
            tu.copy(motion_settings['-Gain'], new_gain)
            data_meta.extend([
                '_rlnMicrographGainName {0}'.format(new_gain.replace(project_folder, '')),
                ])
        else:
            data_meta.extend([
                '_rlnMicrographGainName {0}'.format(new_gain.replace(project_folder, '')),
                ])
            new_gain = None
    else:
        new_gain = None

    if motion_settings['-DefectFile']:
        new_defect = os.path.join(
            project_folder,
            '{0}_defect.mrc'.format(os.path.basename(motion_settings['-DefectFile']))
            )
        if not os.path.exists(new_defect):
            tu.copy(motion_settings['-DefectFile'], new_defect)
            data_meta.extend([
                '_rlnMicrographDefectFile {0}'.format(new_defect.replace(project_folder, '')),
                ])
        else:
            data_meta.extend([
                '_rlnMicrographDefectFile {0}'.format(new_defect.replace(project_folder, '')),
                ])
            new_defect = None
    else:
        new_defect = None

    data_meta.extend([
        '_rlnMicrographBinning {0}'.format(motion_settings['-FtBin']),
        ])
    data_meta.extend([
        '_rlnMicrographOriginalPixelSize {0}'.format(motion_settings['-PixSize']),
        ])
    data_meta.extend([
        '_rlnMicrographDoseRate {0}'.format(motion_settings['-FmDose']),
        ])
    data_meta.extend([
        '_rlnMicrographPreExposure {0}'.format(motion_settings['-InitDose']),
        ])
    data_meta.extend([
        '_rlnVoltage {0}'.format(motion_settings['-kV']),
        ])

    data_meta.extend([
        '_rlnMicrographStartFrame {0}'.format(int(motion_settings['-Throw']) + 1),
        '_rlnMotionModelVersion 0',
        '',
        'data_global_shift',
        '',
        'loop_',
        '_rlnMicrographFrameNumber #1',
        '_rlnMicrographShiftX #2',
        '_rlnMicrographShiftY #3',
        ])

    assert data_original.shape[0] == 1, data_original
    data_original = data_original[0]
    idx_x = 0
    idx_y = 1
    offset_x = data_original[idx_x][0]
    offset_y = data_original[idx_y][0]

    sum_total = 0
    sum_early = 0
    sum_late = 0
    frame_cutoff = float(motion_settings['dose cutoff']) - float(motion_settings['-InitDose'])
    frame_cutoff /= float(motion_settings['-FmDose'])
    for idx in range(1, data_original.shape[1]):
        drift = np.sqrt(
            (data_original[idx_x][idx-1] - data_original[idx_x][idx])**2 +
            (data_original[idx_y][idx-1] - data_original[idx_y][idx])**2
            )
        sum_total += drift
        if idx <= frame_cutoff:
            sum_early += drift
        else:
            sum_late += drift

    for idx in range(data_original.shape[1]):
        data_meta.extend(['{0}\t{1}\t{2}'.format(
            idx+1,
            data_original[idx_x][idx]-offset_x,
            data_original[idx_y][idx]-offset_y
            )])

    relion3_meta = os.path.join(
        log_folder,
        '{0}_transphire_motion_relion3_meta.star'.format(os.path.basename(sum_file))
        )
    with open(relion3_meta, 'w') as write:
        write.write('\n'.join(data_meta))

    header_star_relion3 = co.OrderedDict()
    if dw_file:
        header_star_relion3['_rlnMicrographNameNoDW'] = sum_file.replace(project_folder, '')
        header_star_relion3['_rlnMicrographName'] = dw_file.replace(project_folder, '')
    else:
        header_star_relion3['_rlnMicrographName'] = sum_file.replace(project_folder, '')
    header_star_relion3['_rlnMicrographMetadata'] = relion3_meta.replace(project_folder, '')
    if sum_total:
        header_star_relion3['_rlnAccumMotionTotal'] = sum_total
        header_star_relion3['_rlnAccumMotionEarly'] = sum_early
        header_star_relion3['_rlnAccumMotionLate'] = sum_late
    else:
        pass

    export_lines_star_relion3 = [get_relion_header(header_star_relion3.keys())]
    create_export_data(header_star_relion3.values(), export_lines_star_relion3)

    with open(output_name_star_relion3, 'w') as write:
        write.write('\n'.join(export_lines_star_relion3))

    return output_name_mic_combined, output_name_star_combined, output_name_star_relion3_combined, new_gain, new_defect


def get_relion_header(names):
    """
    Create a relion star file header.

    names - Header names as list

    Returns:
    header string
    """
    header = []
    header.append('')
    header.append('data_')
    header.append('')
    header.append('loop_')
    for index, name in enumerate(names):
        header.append('{0} #{1}'.format(name, index+1))
    return '\n'.join(header)


def create_export_data(export_data, lines):
    """
    Write export data to file.

    export_data - Data to export.
    file_name - Name of the file to write to.

    Returns:
    In place modificaion of lines
    """
    row_string = []
    for value in export_data:
        if isinstance(value, int):
            row_string.append('{0: 7d}'.format(value))
        elif isinstance(value, float):
            row_string.append('{0: 14f}'.format(value))
        else:
            row_string.append('{0:s}'.format(value))
    lines.append('{0}\n'.format('\t'.join(row_string)))



"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
try:
    from PyQt4.QtGui import QWidget, QHBoxLayout, QVBoxLayout, QScrollArea
    from PyQt4.QtCore import pyqtSlot
except ImportError:
    from PyQt5.QtWidgets import QWidget, QHBoxLayout, QVBoxLayout, QScrollArea
    from PyQt5.QtCore import pyqtSlot
from transphire.settingswidget import SettingsWidget
from transphire.separator import Separator
from transphire.tabdocker import TabDocker


class SettingsContainer(QWidget):
    """
    Widget for setting widgets.

    Inherits:
    QWidget
    """

    def __init__(self, content, name, parent=None, **kwargs):
        """
        Initialise layout of the widget.

        Arguments:
        content - Content for the widget
        max_widgets - Maximum widgets per column
        parent - Parent widget (default None)

        Returns:
        None
        """
        super(SettingsContainer, self).__init__(parent)

        # TabDocker widget for Main and Advanced
        my_tab_docker = TabDocker(self)
        my_tab_docker.setTabPosition('South')
        layout_main = QVBoxLayout(self)
        layout_main.setContentsMargins(0, 0, 0, 0)
        layout_main.addWidget(my_tab_docker)

        self.layout_dict = {}
        for entry in content:
            for widget in entry:
                for key in widget:
                    if key == 'WIDGETS MAIN':
                        self.layout_dict['Main_max'] = int(widget['WIDGETS MAIN'][0])
                    elif key == 'WIDGETS ADVANCED':
                        self.layout_dict['Advanced_max'] = int(widget['WIDGETS ADVANCED'][0])
                    elif key == 'WIDGETS RARE':
                        self.layout_dict['Rare_max'] = int(widget['WIDGETS RARE'][0])
                    else:
                        continue

        for dict_name in ['Main', 'Advanced', 'Rare']:
            widget = QWidget(self)
            widget.setObjectName('settings')

            self.layout_dict[dict_name] = QHBoxLayout(widget)
            self.layout_dict[dict_name].setContentsMargins(3, 3, 3, 3)
            self.layout_dict[dict_name].addStretch(1)

            self.layout_dict['{0}_v'.format(dict_name)] = None
            self.layout_dict['{0}_idx'.format(dict_name)] = 0

            scroll_area = QScrollArea(my_tab_docker)
            scroll_area.setWidgetResizable(True)
            scroll_area.setWidget(widget)
            my_tab_docker.add_tab(scroll_area, dict_name)

        # Global content
        self.content = {}
        self.group = {}

        # Add to layout
        for entry in content:
            for widget in entry:
                for key in widget:
                    if key == 'WIDGETS MAIN' or key == 'WIDGETS ADVANCED' or key == 'WIDGETS RARE':
                        continue
                    layout_name = widget[key][1]['widget_2']
                    widget_name = widget[key][1]['name']
                    group = widget[key][1]['group']

                    if self.layout_dict['{0}_idx'.format(layout_name)] % self.layout_dict['{0}_max'.format(layout_name)] == 0:
                        if self.layout_dict['{0}_v'.format(layout_name)] is not None:
                            self.layout_dict[layout_name].addWidget(Separator(typ='vertical', color='lightgrey'))
                            self.layout_dict['{0}_v'.format(layout_name)].addStretch(1)
                        else:
                            pass
                        self.layout_dict['{0}_v'.format(layout_name)] = QVBoxLayout()
                        self.layout_dict['{0}_v'.format(layout_name)].setContentsMargins(0, 0, 0, 0)
                        self.layout_dict[layout_name].addLayout(self.layout_dict['{0}_v'.format(layout_name)])

                    widget = SettingsWidget(content=widget[key], name=name, parent=self)
                    if group:
                        group, state = group.split(':')
                        self.group.setdefault(group, [])
                        self.group[group].append([widget, state, widget_name])
                    self.content[widget_name] = widget
                    self.layout_dict['{0}_v'.format(layout_name)].addWidget(widget)
                    self.layout_dict['{0}_idx'.format(layout_name)] += 1

        for key in self.group:
            self.content[key].sig_index_changed.connect(self.change_state)
            self.change_state(name=key)

        for dict_name in ['Main', 'Advanced', 'Rare']:
            try:
                self.layout_dict['{0}_v'.format(dict_name)].addStretch(1)
                self.layout_dict[dict_name].addStretch(1)
            except AttributeError:
                pass

    @pyqtSlot(str)
    def change_state(self, name):
        """
        Change the state of widgets based on the choice of the corresponding combo box

        name - Name of the group to change status (Emitted by the combo box)

        Returns:
        None
        """
        try:
            for entry in self.group[name]:
                widget = entry[0]
                state = entry[1]
                sub_name = entry[2]
                if not self.content[name].isEnabled():
                    widget.setEnabled(False)
                elif self.content[name].edit.currentText() == state:
                    widget.setEnabled(True)
                else:
                    widget.setEnabled(False)
                self.change_state(name=sub_name)
        except KeyError:
            return None

    def get_settings(self, quiet=False):
        """
        Get the settings as dict

        Arguments:
        quiet - If True, no prints are executed

        Returns:
        Settings as dictionary
        """
        settings = {}
        error = False
        for key in self.content:
            dictionary = self.content[key].get_settings(quiet=quiet)
            if dictionary is None:
                error = True
            else:
                settings.update(dictionary)
            if key == 'Motion' or key == 'CTF' or key == 'Picking':
                new_key = '{0}_entries'.format(key)
                settings[new_key] = self.content[key].get_combo_entries()
            else:
                pass
        if error:
            return None
        else:
            return [settings]

    def set_settings(self, settings):
        """
        Set settings to the widgets

        Arguments:
        settings - Settings to set.

        Returns:
        None
        """
        for key in settings:
            try:
                content = self.content[key]
            except KeyError:
                if key == 'CTF_entries':
                    continue
                elif key == 'Motion_entries':
                    continue
                elif key == 'Picking_entries':
                    continue
                else:
                    print('Content for {0} is disabled.'.format(key))
                    continue

            try:
                content.set_settings(settings[key])
            except KeyError:
                if key == 'CTF_entries':
                    pass
                elif key == 'Motion_entries':
                    pass
                elif key == 'Picking_entries':
                    pass
                else:
                    print(
                        'Setting changed: {0}!'.format(key),
                        'Please do provide new values'
                        )

    def enable(self, var, use_all):
        """
        Disable or enable all widgets

        Arguments:
        var - If True, enable the widgets, else disable them
        use_all - If True, all widgets are enabled/disabled

        Returns:
        None
        """
        for key in self.content:
            if use_all:
                self.content[key].setEnabled(var)
            else:
                self.content[key].setEnabled(var)

        if var:
            try:
                self.set_settings(self.get_settings(quiet=True)[0])
            except TypeError:
                pass
        else:
            pass



import pkg_resources
from pkg_resources import get_distribution
import os
import re

here = os.path.normcase(__file__)
try:
    _dist = get_distribution('transphire')
    # Normalize case for Windows systems
    dist_loc = os.path.normcase(_dist.location)
    if not here.startswith(os.path.join(dist_loc, 'transphire')):
        # not installed, but there is another version that *is*
        raise pkg_resources.DistributionNotFound
except pkg_resources.DistributionNotFound:
    try:
        pip_path = '{0}/../../../../../bin/pip'.format(os.path.dirname(here))
        pip_output = os.popen('{0} freeze'.format(pip_path)).read()
        transphire = re.search(r'(transphire==[0-9\.]*)', pip_output)
        __version__ = transphire.group(1).split('==')[1]
    except:
        __version__ = 'XX.XX.XX'
        print('Could not find version number! Please install this project with setup.py!')
        print('Use: "pip freeze" to find the transphire version')
else:
    __version__ = _dist.version



"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
try:
    from PyQt4.QtGui import QWidget, QVBoxLayout, QTabWidget
except ImportError:
    from PyQt5.QtWidgets import QWidget, QVBoxLayout, QTabWidget


class TabDocker(QWidget):
    """
    Tab widget for the settingswidgets.

    Inherits:
    QWidget
    """

    def __init__(self, parent=None, **kwargs):
        """
        Initialise layout for TabDocker

        Arguments:
        parent - Parent widget (default None)

        Return:
        None
        """
        super(TabDocker, self).__init__(parent)
        layout_tmp = QVBoxLayout(self)
        self.parent_widget = QWidget(self)
        layout_tmp.addWidget(self.parent_widget)
        layout_tmp.setContentsMargins(0, 0, 0, 0)

        layout = QVBoxLayout(self.parent_widget)
        self.tab_widget = QTabWidget(self)
        layout.addWidget(self.tab_widget)

    def add_tab(self, widget, name):
        """
        Add a new tab to the TabDocker

        Arguments:
        widget - Widget to add
        name - Name of the widget

        Return:
        None
        """
        if isinstance(widget, TabDocker):
            widget.parent_widget.setObjectName('tab')
        else:
            pass
        self.tab_widget.addTab(widget, name)

    def count(self):
        """
        Return the number of tabs.

        Arguments:
        None

        Returns:
        Number of tabs
        """
        return self.tab_widget.count()

    def widget(self, idx):
        """
        Return the widget that belongs to the idx of tabs.

        Arguments:
        idx - Tab index

        Returns:
        Widget
        """
        return self.tab_widget.widget(idx)

    def setMovable(self, status):
        """
        Set the movable status for the tab widgets

        Arguments:
        status - Boolean variable for the status

        Returns:
        None
        """
        return self.tab_widget.setMovable(status)

    def tabText(self, idx):
        """
        Return the text of the tab at idx

        Arguments:
        idx - Index of the tab

        Returns:
        Text of the tab at position isx
        """
        return self.tab_widget.tabText(idx)

    def setTabText(self, idx, text):
        """
        Set the text for the tab at idx

        Arguments:
        idx - Index of the tab
        text - Text of the tab

        Returns:
        None
        """
        return self.tab_widget.setTabText(idx, text)

    def removeTab(self, idx):
        """
        Remove the widget located at tab idx

        Arguments:
        idx - Idx of the widget

        Returns:
        None
        """
        return self.tab_widget.removeTab(idx)

    def indexOf(self, widget):
        """
        Get the index of the widget.

        Arguments:
        widget - Adress of the widget

        Returns:
        Index of the widget
        """
        return self.tab_widget.indexOf(widget)

    def setTabPosition(self, position):
        """
        Set the tab position of the Tab bar

        Arguments:
        position - Tab position as string ['North', 'East', 'West', 'South']

        Returns:
        None
        """
        tab_position_dict = {
            'North': QTabWidget.North,
            'South': QTabWidget.South,
            'West': QTabWidget.West,
            'East': QTabWidget.East,
            }
        self.tab_widget.setTabPosition(tab_position_dict[position])



"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import warnings
import imageio
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_qt4agg import (
    FigureCanvasQTAgg as FigureCanvas,
    NavigationToolbar2QT as NavigationToolbar
    )
try:
    from PyQt4.QtGui import QWidget, QVBoxLayout, QHBoxLayout, QComboBox, QPushButton
    from PyQt4.QtCore import pyqtSlot
except ImportError:
    from PyQt5.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QComboBox, QPushButton
    from PyQt5.QtCore import pyqtSlot
from transphire import transphire_utils as tu
warnings.filterwarnings('ignore')


class PlotWidget(QWidget):
    """
    PlotWidget widget.
    Widget used to show data plots.

    Inherits from:
    QWidget

    Signals:
    None
    """

    def __init__(self, label, plot_typ, *args, parent=None, **kwargs):
        """
        Setup the layout for the widget.

        Arguments:
        label - Label of the plot.
        plot_typ - Type of plot (e.g. histogram, values)
        parent - Parent widget (Default None)
        *args - Unused additional arguments
        **kwargs - Unused named additional arguments

        Return:
        None
        """
        super(PlotWidget, self).__init__(parent)

        # Variables
        self.label = label
        self.plot_typ = plot_typ

        # Initialise figure
        self.figure = plt.figure()
        self.ax1 = self.figure.add_subplot(111)
        self.compute_initial_figure()

        # Create canvas for the figure
        self.canvas = FigureCanvas(self.figure)
        self.canvas.setParent(self)
        self.toolbar = NavigationToolbar(self.canvas, self)
        self.canvas.draw_idle()

        layout_v = QVBoxLayout(self)
        # Add for image plot typ
        if self.plot_typ == 'image' and self.label == 'image':
            self.data = np.array([])
            self.idx = 0
            self.default_value = 'latest'

            layout_h = QHBoxLayout()
            self.combo_box = QComboBox(self)
            self.combo_box.addItems([self.default_value])
            self.combo_box.currentIndexChanged.connect(lambda: self.change_idx('combo'))
            layout_h.addWidget(self.combo_box)
            self.prev_button = QPushButton('Prev', self)
            self.prev_button.setObjectName('frame')
            self.prev_button.clicked.connect(lambda: self.change_idx('prev'))
            self.prev_button.setEnabled(False)
            layout_h.addWidget(self.prev_button)
            self.next_button = QPushButton('Next', self)
            self.next_button.setObjectName('frame')
            self.next_button.clicked.connect(lambda: self.change_idx('next'))
            self.next_button.setEnabled(False)
            layout_h.addWidget(self.next_button)
            self.reset_button = QPushButton(self.default_value, self)
            self.reset_button.setObjectName('frame')
            self.reset_button.clicked.connect(lambda: self.change_idx('reset'))
            layout_h.addWidget(self.reset_button)
            layout_h.addStretch(1)
            layout_v.addLayout(layout_h)

        # Fill layout
        layout_v.addWidget(self.toolbar)
        layout_v.addWidget(self.canvas)


    def compute_corrupted_figure(self, title):
        """
        Compute the corrupted figure shown on startup.

        Arguments:
        None

        Return:
        None
        """
        self.ax1.clear()
        self.ax1.plot([0.5, 0, 0, 0.5], [1, 1, 0, 0], '#68a3c3')
        self.ax1.plot([1, 1.5, 1.5, 1, 1], [0, 0, 1, 1, 0], '#68a3c3')
        self.ax1.plot([2, 2, 2.5, 2.5, 2, 2.5], [0, 1, 1, 0.5, 0.5, 0], '#68a3c3')
        self.ax1.plot([3, 3, 3.5, 3.5, 3, 3.5], [0, 1, 1, 0.5, 0.5, 0], '#68a3c3')
        self.ax1.plot([4, 4, 4.5, 4.5], [1, 0, 0, 1], '#68a3c3')
        self.ax1.plot([5, 5, 5.5, 5.5, 5], [0, 1, 1, 0.5, 0.5], '#68a3c3')
        self.ax1.plot([6, 6.5, 6.25, 6.25], [1, 1, 1, 0], '#68a3c3')
        self.ax1.set_title(title)


    def compute_initial_figure(self):
        """
        Compute the initial figure shown on startup.

        Arguments:
        None

        Return:
        None
        """
        self.ax1.clear()
        self.ax1.plot([0, 0, 0, 0.5, 0.5, 0.5], [0, 1, 0.5, 0.5, 0, 1], '#68a3c3')
        self.ax1.plot([1.5, 1, 1, 1.5, 1, 1, 1.5], [0, 0, 0.5, 0.5, 0.5, 1, 1], '#68a3c3')
        self.ax1.plot([2, 2, 2.5], [1, 0, 0], '#68a3c3')
        self.ax1.plot([3, 3, 3.5], [1, 0, 0], '#68a3c3')
        self.ax1.plot([4, 4.15, 4.35, 4.5, 4.5, 4.35, 4.15, 4, 4], [0.3, 0, 0, 0.3, 0.6, 1, 1, 0.6, 0.3], '#68a3c3')


    @pyqtSlot(str, object, str, object)
    def update_figure(self, name, data, directory_name, settings):
        """
        Update the figure with data plot.

        Arguments:
        name - Name of the plot type
        data - Data to plot
        directory_name - Name of the output directory
        settings - Settings provided by the user

        Return:
        None
        """
        if name == 'False':
            return None
        elif name == 'Later':
            return None
        elif self.plot_typ == 'values' or self.plot_typ == 'histogram':
            x_values, y_values, label, title = tu.get_function_dict()[name]['plot'](
                data=data,
                settings=settings,
                label=self.label,
                )
            idx_nan = np.isnan(y_values)
            x_values = x_values[~idx_nan]
            y_values = y_values[~idx_nan]

            color = '#68a3c3'

            self.ax1.clear()
            title_list = [title]
            if idx_nan.any():
                title_list.append(
                    '{0} entries not valid due to NAN'.format(idx_nan.sum())
                    )

            if self.plot_typ == 'values':
                x_label = 'Micrograph ID'
                y_label = label
                self.ax1.plot(x_values, y_values, '.', color=color)
            elif self.plot_typ == 'histogram':
                x_label = label
                y_label = 'Nr. of micrographs'
                if np.max(y_values) - np.min(y_values) < 0.001:
                    self.ax1.hist(y_values, 1)
                    self.ax1.set_xlim([np.max(y_values)-1, np.min(y_values)+1])
                elif title == 'Resolution limit':
                    if np.max(y_values) > 20:
                        self.ax1.hist(y_values[y_values <= 20], 100, color=color)
                        title_list.append('{0} micrographs out of range (0,20)'.format(
                            y_values[y_values > 20].shape[0]
                            ))
                    else:
                        self.ax1.hist(y_values, 100, color=color)
                else:
                    self.ax1.hist(y_values, 100, color=color)
            else:
                print('Plotwidget - ', self.plot_typ, ' is not known!!!')
                return None

            self.ax1.grid()
            self.ax1.set_title('\n'.join(title_list))
            self.ax1.set_xlabel(x_label)
            self.ax1.set_ylabel(y_label)
            try:
                self.canvas.draw_idle()
            except RecursionError:
                import sys
                print('sys.getrecursionlimit()')
                print(sys.getrecursionlimit())
                print('self.plot_typ')
                print(self.plot_typ)
                print('label')
                print(label)
                print('x_label')
                print(x_label)
                print('y_label')
                print(y_label)
                print('directory_name')
                print(directory_name)
                print('title')
                print(title)
                print('x_values')
                print(x_values)
                print('len x_values')
                print(len(x_values))
                print('y_values')
                print(y_values)
                print('len y_values')
                print(len(y_values))
                print('COULD NOT DRAW!!!')
                print('Please contact the TranSHPIRE authors!!!')
                print('Restarting TranSPHIRE will fix this issue.')
                return None

            output_name = os.path.join(
                directory_name,
                '{0}_{1}.png'.format(
                    self.label,
                    self.plot_typ
                    )
                )
            try:
                self.figure.savefig(output_name.replace(' ', '_'))
            except RuntimeError:
                pass
            except FileNotFoundError:
                pass

        elif self.plot_typ == 'image':
            self.data = data[::-1]
            current_text = self.combo_box.currentText()

            self.combo_box.blockSignals(True)
            self.combo_box.clear()
            files = [self.default_value]
            files.extend(self.data['file_name'].tolist())
            self.combo_box.addItems(files)

            self.idx = max(0, self.combo_box.findText(current_text))
            self.combo_box.setCurrentIndex(self.idx)
            self.combo_box.blockSignals(False)

            if self.idx <= 0 and self.idx >= self.data.shape[0]:
                self.prev_button.setEnabled(False)
                self.next_button.setEnabled(False)
            elif self.idx <= 0:
                self.next_button.setEnabled(False)
                self.prev_button.setEnabled(True)
            elif self.idx >= self.data.shape[0]:
                self.next_button.setEnabled(True)
                self.prev_button.setEnabled(False)
            else:
                self.next_button.setEnabled(True)
                self.prev_button.setEnabled(True)

            if str(current_text) == self.default_value:
                self.show_image()
            else:
                pass

        else:
            print('Plotwidget - ', self.plot_typ, ' is not known!!!')
            return None

    @pyqtSlot()
    def show_image(self):
        if self.data.shape[0] == 0:
            return None
        else:
            pass

        if str(self.combo_box.currentText()) == self.default_value:
            idx = 0
        else:
            idx = self.idx-1

        self.ax1.clear()
        if self.data['object'][idx] is None:
            try:
                jpg_data = imageio.imread(self.data['image'][idx])
            except Exception as e:
                print('Error loading image: {0} -- Message: {1}'.format(self.data['image'][idx], str(e)))
                self.compute_corrupted_figure(title=self.data['file_name'][idx])
            else:
                self.ax1.imshow(jpg_data)
        else:
                self.ax1.imshow(self.data['object'][idx])
        self.ax1.get_xaxis().set_visible(False)
        self.ax1.get_yaxis().set_visible(False)
        self.ax1.set_title(self.data['file_name'][idx])
        self.canvas.draw_idle()

    def change_idx(self, typ):
        if typ == 'next':
            if self.idx <= 0 and self.idx >= self.data.shape[0]:
                self.prev_button.setEnabled(False)
                self.next_button.setEnabled(False)
            elif self.idx <= 0:
                self.next_button.setEnabled(False)
            else:
                self.prev_button.setEnabled(True)
                self.idx -= 1
        elif typ == 'prev':
            if self.idx <= 0 and self.idx >= self.data.shape[0]:
                self.prev_button.setEnabled(False)
                self.next_button.setEnabled(False)
            elif self.idx >= self.data.shape[0]:
                self.prev_button.setEnabled(False)
            else:
                self.next_button.setEnabled(True)
                self.idx += 1
        elif typ == 'combo':
            self.idx = self.combo_box.currentIndex()
            if self.idx <= 0 and self.idx >= self.data.shape[0]:
                self.idx = 0
                self.prev_button.setEnabled(False)
                self.next_button.setEnabled(False)
            elif self.idx <= 0:
                self.idx = 0
                self.next_button.setEnabled(False)
                self.prev_button.setEnabled(True)
            elif self.idx >= self.data.shape[0]:
                self.next_button.setEnabled(True)
                self.prev_button.setEnabled(False)
            else:
                self.next_button.setEnabled(True)
                self.next_button.setEnabled(True)

        elif typ == 'reset':
            self.idx = 0
            if self.idx <= 0 and self.idx >= self.data.shape[0]:
                self.prev_button.setEnabled(False)
                self.next_button.setEnabled(False)
            elif self.idx <= 0:
                self.next_button.setEnabled(False)
                self.prev_button.setEnabled(True)
            elif self.idx >= self.data.shape[0]:
                self.next_button.setEnabled(True)
                self.prev_button.setEnabled(False)
            else:
                self.next_button.setEnabled(True)
                self.next_button.setEnabled(True)
        else:
            pass
        self.combo_box.setCurrentIndex(self.idx)
        self.show_image()
